@@ -10,10 +10,18 @@
 #ifndef _WIN32
 #include <sys/uio.h>
 #include <poll.h>
+#else
+#include <io.h>
 #endif
 
 #include "processx.h"
 
+#ifdef _WIN32
+#include "win/processx-win.h"
+#else
+#include "unix/processx-unix.h"
+#endif
+
 /* Internal functions in this file */
 
 static void processx__connection_find_chars(processx_connection_t *ccon,
@@ -74,6 +82,69 @@ SEXP processx_connection_create(SEXP handle, SEXP encoding) {
   return result;
 }
 
+SEXP processx_connection_create_fd(SEXP handle, SEXP encoding, SEXP close) {
+  int fd = INTEGER(handle)[0];
+  const char *c_encoding = CHAR(STRING_ELT(encoding, 0));
+  processx_file_handle_t os_handle;
+  processx_connection_t *con;
+  SEXP result = R_NilValue;
+
+#ifdef _WIN32
+  os_handle = (HANDLE) _get_osfhandle(fd);
+#else
+  os_handle = fd;
+#endif
+
+  con = processx_c_connection_create(os_handle, PROCESSX_FILE_TYPE_ASYNCPIPE,
+				     c_encoding, &result);
+
+  if (! LOGICAL(close)[0]) con->close_on_destroy = 0;
+
+  return result;
+}
+
+SEXP processx_connection_create_file(SEXP filename, SEXP read, SEXP write) {
+  const char *c_filename = CHAR(STRING_ELT(filename, 0));
+  int c_read = LOGICAL(read)[0];
+  int c_write = LOGICAL(write)[0];
+  SEXP result = R_NilValue;
+  processx_file_handle_t os_handle;
+
+#ifdef _WIN32
+  DWORD access = 0, create = 0;
+  if (c_read) access |= GENERIC_READ;
+  if (c_write) access |= GENERIC_WRITE;
+  if (c_read) create |= OPEN_EXISTING;
+  if (c_write) create |= CREATE_ALWAYS;
+  os_handle = CreateFile(
+    /* lpFilename = */ c_filename,
+    /* dwDesiredAccess = */ access,
+    /* dwShareMode = */ 0,
+    /* lpSecurityAttributes = */ NULL,
+    /* dwCreationDisposition = */ create,
+    /* dwFlagsAndAttributes = */ FILE_ATTRIBUTE_NORMAL,
+    /* hTemplateFile = */ NULL);
+  if (os_handle == INVALID_HANDLE_VALUE) {
+    PROCESSX_ERROR("Cannot open file", GetLastError());
+  }
+
+#else
+  int flags = 0;
+  if ( c_read && !c_write) flags |= O_RDONLY;
+  if (!c_read &&  c_write) flags |= O_WRONLY | O_CREAT | O_TRUNC;
+  if ( c_read &&  c_write) flags |= O_RDWR;
+  os_handle = open(c_filename, flags, 0644);
+  if (os_handle == -1) {
+    error("Cannot open file `%s`: `%s`", c_filename, strerror(errno));
+  }
+#endif
+
+  processx_c_connection_create(os_handle, PROCESSX_FILE_TYPE_FILE,
+			       "", &result);
+
+  return result;
+}
+
 SEXP processx_connection_read_chars(SEXP con, SEXP nchars) {
 
   processx_connection_t *ccon = R_ExternalPtrAddr(con);
@@ -133,6 +204,22 @@ SEXP processx_connection_read_lines(SEXP con, SEXP nlines) {
   return result;
 }
 
+SEXP processx_connection_write_bytes(SEXP con, SEXP bytes) {
+  processx_connection_t *ccon = R_ExternalPtrAddr(con);
+  Rbyte *cbytes = RAW(bytes);
+  size_t nbytes = LENGTH(bytes);
+  SEXP result;
+
+  ssize_t written = processx_c_connection_write_bytes(ccon, cbytes, nbytes);
+
+  size_t left = nbytes - written;
+  PROTECT(result = allocVector(RAWSXP, left));
+  if (left > 0) memcpy(RAW(result), cbytes + written, left);
+
+  UNPROTECT(1);
+  return result;
+}
+
 SEXP processx_connection_is_eof(SEXP con) {
   processx_connection_t *ccon = R_ExternalPtrAddr(con);
   if (!ccon) error("Invalid connection object");
@@ -159,6 +246,156 @@ SEXP processx_connection_poll(SEXP pollables, SEXP timeout) {
   return R_NilValue;
 }
 
+SEXP processx_connection_create_pipepair(SEXP encoding) {
+  const char *c_encoding = CHAR(STRING_ELT(encoding, 0));
+  SEXP result, con1, con2;
+
+#ifdef _WIN32
+  HANDLE h1, h2;
+  processx__create_pipe(0, &h1, &h2);
+
+#else
+  int pipe[2], h1, h2;
+  processx__make_socketpair(pipe);
+  processx__nonblock_fcntl(pipe[0], 1);
+  processx__nonblock_fcntl(pipe[1], 0);
+  h1 = pipe[0];
+  h2 = pipe[1];
+#endif
+
+  processx_c_connection_create(h1, PROCESSX_FILE_TYPE_ASYNCPIPE,
+			       c_encoding, &con1);
+  PROTECT(con1);
+  processx_c_connection_create(h2, PROCESSX_FILE_TYPE_ASYNCPIPE,
+			       c_encoding, &con2);
+  PROTECT(con2);
+
+  result = PROTECT(allocVector(VECSXP, 2));
+  SET_VECTOR_ELT(result, 0, con1);
+  SET_VECTOR_ELT(result, 1, con2);
+
+  UNPROTECT(3);
+  return result;
+}
+
+SEXP processx__connection_set_std(SEXP con, int which, int drop) {
+  processx_connection_t *ccon = R_ExternalPtrAddr(con);
+  if (!ccon) error("Invalid connection object");
+  SEXP result = R_NilValue;
+
+#ifdef _WIN32
+  int fd, ret;
+  if (!drop) {
+    int saved = _dup(which);
+    processx_file_handle_t os_handle;
+    if (saved == -1) {
+      PROCESSX_ERROR("Cannot save stdout/stderr for rerouting", GetLastError());
+    }
+    os_handle = (HANDLE) _get_osfhandle(saved) ;
+    processx_c_connection_create(os_handle, PROCESSX_FILE_TYPE_PIPE,
+				 "", &result);
+  }
+  fd = _open_osfhandle((intptr_t) ccon->handle.handle, 0);
+  ret = _dup2(fd, which);
+  if (ret) PROCESSX_ERROR("Cannot reroute stdout/stderr", GetLastError());
+
+#else
+  const char *what[] = { "stdin", "stdout", "stderr" };
+  int ret;
+  if (!drop) {
+    processx_file_handle_t os_handle = dup(which);
+    if (os_handle == -1) {
+      error("Cannot save %s for rerouting: `%s`", what[which], strerror(errno));
+    }
+    processx_c_connection_create(os_handle, PROCESSX_FILE_TYPE_PIPE,
+				 "", &result);
+  }
+  ret = dup2(ccon->handle, which);
+  if (ret == -1) {
+    error("Cannot reroute %s: `%s`", what[which], strerror(errno));
+  }
+#endif
+
+  return result;
+}
+
+SEXP processx_connection_set_stdout(SEXP con, SEXP drop) {
+  return processx__connection_set_std(con, 1, LOGICAL(drop)[0]);
+}
+
+SEXP processx_connection_set_stderr(SEXP con, SEXP drop) {
+  return processx__connection_set_std(con, 2, LOGICAL(drop)[0]);
+}
+
+SEXP processx_connection_get_fileno(SEXP con) {
+  processx_connection_t *ccon = R_ExternalPtrAddr(con);
+  if (!ccon) error("Invalid connection object");
+  int fd;
+
+#ifdef _WIN32
+  fd = _open_osfhandle((intptr_t) ccon->handle.handle, 0);
+#else
+  fd = ccon->handle;
+#endif
+
+  return ScalarInteger(fd);
+}
+
+#ifdef _WIN32
+
+/*
+ * Clear the HANDLE_FLAG_INHERIT flag from all HANDLEs that were inherited
+ * the parent process. Don't check for errors - the stdio handles may not be
+ * valid, or may be closed already. There is no guarantee that this function
+ * does a perfect job.
+ */
+
+SEXP processx_connection_disable_inheritance() {
+  HANDLE handle;
+  STARTUPINFOW si;
+
+  /* Make the windows stdio handles non-inheritable. */
+  handle = GetStdHandle(STD_INPUT_HANDLE);
+  if (handle != NULL && handle != INVALID_HANDLE_VALUE) {
+    SetHandleInformation(handle, HANDLE_FLAG_INHERIT, 0);
+  }
+
+  handle = GetStdHandle(STD_OUTPUT_HANDLE);
+  if (handle != NULL && handle != INVALID_HANDLE_VALUE) {
+    SetHandleInformation(handle, HANDLE_FLAG_INHERIT, 0);
+  }
+
+  handle = GetStdHandle(STD_ERROR_HANDLE);
+  if (handle != NULL && handle != INVALID_HANDLE_VALUE) {
+    SetHandleInformation(handle, HANDLE_FLAG_INHERIT, 0);
+  }
+
+  /* Make inherited CRT FDs non-inheritable. */
+  GetStartupInfoW(&si);
+  if (processx__stdio_verify(si.lpReserved2, si.cbReserved2)) {
+    processx__stdio_noinherit(si.lpReserved2);
+  }
+
+  return R_NilValue;
+}
+
+#else
+
+SEXP processx_connection_disable_inheritance() {
+  int fd;
+
+  /* Set the CLOEXEC flag on all open descriptors. Unconditionally try the
+   * first 16 file descriptors. After that, bail out after the first error.
+   */
+  for (fd = 0; ; fd++) {
+    if (processx__cloexec_fcntl(fd, 1) && fd > 15) break;
+  }
+
+  return R_NilValue;
+}
+
+#endif
+
 /* Api from C -----------------------------------------------------------*/
 
 processx_connection_t *processx_c_connection_create(
@@ -177,6 +414,7 @@ processx_connection_t *processx_c_connection_create(
   con->is_closed_ = 0;
   con->is_eof_  = 0;
   con->is_eof_raw_ = 0;
+  con->close_on_destroy = 1;
   con->iconv_ctx = 0;
 
   con->buffer = 0;
@@ -201,55 +439,55 @@ processx_connection_t *processx_c_connection_create(
   con->handle.handle = os_handle;
   memset(&con->handle.overlapped, 0, sizeof(OVERLAPPED));
   con->handle.read_pending = FALSE;
-  con->handle.overlapped.hEvent = CreateEvent(
-    /* lpEventAttributes = */ NULL,
-    /* bManualReset = */      FALSE,
-    /* bInitialState = */     FALSE,
-    /* lpName = */            NULL);
-
-  if (con->handle.overlapped.hEvent == NULL) {
-    free(con);
-    PROCESSX_ERROR("Cannot create connection event", GetLastError());
-    return 0; 			/* never reached */
-  }
-
-  HANDLE iocp = processx__get_default_iocp();
-  HANDLE res = CreateIoCompletionPort(
-    /* FileHandle =  */                con->handle.handle,
-    /* ExistingCompletionPort = */     iocp,
-    /* CompletionKey = */              (ULONG_PTR) con,
-    /* NumberOfConcurrentThreads = */  0);
-
-  if (!res) PROCESSX_ERROR("cannot add file to IOCP", GetLastError());
-
+  con->handle.freelist = FALSE;
 #else
   con->handle = os_handle;
 #endif
 
   if (r_connection) {
     result = PROTECT(R_MakeExternalPtr(con, R_NilValue, R_NilValue));
-    R_RegisterCFinalizerEx(result, processx__connection_xfinalizer, 1);
+    R_RegisterCFinalizerEx(result, processx__connection_xfinalizer, 0);
     class = PROTECT(ScalarString(mkChar("processx_connection")));
     setAttrib(result, R_ClassSymbol, class);
     *r_connection = result;
+    UNPROTECT(2);
   }
 
-  if (r_connection) UNPROTECT(2);
   return con;
 }
 
 /* Destroy */
 void processx_c_connection_destroy(processx_connection_t *ccon) {
 
-  processx_c_connection_close(ccon);
-
   if (!ccon) return;
 
-  if (ccon->iconv_ctx) Riconv_close(ccon->iconv_ctx);
+  if (ccon->close_on_destroy) processx_c_connection_close(ccon);
 
-  if (ccon->buffer) free(ccon->buffer);
-  if (ccon->utf8) free(ccon->utf8);
-  if (ccon->encoding) free(ccon->encoding);
+  /* Even if not close_on_destroy, for us the connection is closed. */
+  ccon->is_closed_ = 1;
+
+#ifdef _WIN32
+  /* Check if we can free the connection. If there is a pending read,
+     then we cannot. In this case schedule_destroy will add it to a free
+     list and return 1. */
+  if (processx__connection_schedule_destroy(ccon)) return;
+#endif
+
+  if (ccon->iconv_ctx) {
+    Riconv_close(ccon->iconv_ctx);
+    ccon->iconv_ctx = NULL;
+  }
+
+  if (ccon->buffer) { free(ccon->buffer); ccon->buffer = NULL; }
+  if (ccon->utf8) { free(ccon->utf8); ccon->utf8 = NULL; }
+  if (ccon->encoding) { free(ccon->encoding); ccon->encoding = NULL; }
+
+ #ifdef _WIN32
+  if (ccon->handle.overlapped.hEvent) {
+    CloseHandle(ccon->handle.overlapped.hEvent);
+  }
+  ccon->handle.overlapped.hEvent = 0;
+#endif
 
   free(ccon);
 }
@@ -344,6 +582,38 @@ ssize_t processx_c_connection_read_line(processx_connection_t *ccon,
   return newline;
 }
 
+/* Write bytes */
+ssize_t processx_c_connection_write_bytes(
+  processx_connection_t *ccon,
+  const void *buffer,
+  size_t nbytes) {
+
+  PROCESSX_CHECK_VALID_CONN(ccon);
+
+#ifdef _WIN32
+  DWORD written;
+  BOOL ret = WriteFile(
+    /* hFile =                  */ ccon->handle.handle,
+    /* lpBuffer =               */ buffer,
+    /* nNumberOfBytesToWrite =  */ nbytes,
+    /* lpNumberOfBytesWritten = */ &written,
+    /* lpOverlapped =           */ NULL);
+  if (!ret) PROCESSX_ERROR("Cannot write connection ", GetLastError());
+  return (ssize_t) written;
+#else
+  ssize_t ret = write(ccon->handle, buffer, nbytes);
+  if (ret == -1) {
+    if (errno == EAGAIN || errno == EWOULDBLOCK) {
+      return 0;
+    } else {
+      error("Cannot write connection: %s at %s:%d", strerror(errno),
+	    __FILE__, __LINE__);
+    }
+  }
+  return ret;
+#endif
+}
+
 /* Check if the connection has ended */
 int processx_c_connection_is_eof(processx_connection_t *ccon) {
   return ccon->is_eof_;
@@ -353,14 +623,9 @@ int processx_c_connection_is_eof(processx_connection_t *ccon) {
 void processx_c_connection_close(processx_connection_t *ccon) {
 #ifdef _WIN32
   if (ccon->handle.handle) {
-    CancelIo(ccon->handle.handle);
     CloseHandle(ccon->handle.handle);
   }
   ccon->handle.handle = 0;
-  if (ccon->handle.overlapped.hEvent) {
-    CloseHandle(ccon->handle.overlapped.hEvent);
-  }
-  ccon->handle.overlapped.hEvent = 0;
 #else
   if (ccon->handle >= 0) close(ccon->handle);
   ccon->handle = -1;
@@ -374,41 +639,118 @@ int processx_c_connection_is_closed(processx_connection_t *ccon) {
 
 #ifdef _WIN32
 
+/* TODO: errors */
+int processx__socket_pair(SOCKET fds[2]) {
+  struct sockaddr_in inaddr;
+  struct sockaddr addr;
+  SOCKET lst = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
+  memset(&inaddr, 0, sizeof(inaddr));
+  memset(&addr, 0, sizeof(addr));
+  inaddr.sin_family = AF_INET;
+  inaddr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
+  inaddr.sin_port = 0;
+  int yes = 1;
+  setsockopt(lst, SOL_SOCKET, SO_REUSEADDR, (char*)&yes, sizeof(yes));
+  bind(lst, (struct sockaddr *)&inaddr, sizeof(inaddr));
+  listen(lst, 1);
+  int len = sizeof(inaddr);
+  getsockname(lst, &addr,&len);
+  fds[0] = socket(AF_INET, SOCK_STREAM, 0);
+  connect(fds[0], &addr, len);
+  fds[1] = accept(lst,0,0);
+  closesocket(lst);
+  return 0;
+}
+
 int processx_c_connection_poll(processx_pollable_t pollables[],
 			       size_t npollables, int timeout) {
 
   int hasdata = 0;
-  size_t i, j = 0;
+  size_t i, j = 0, selj = 0;
   int *ptr;
   int timeleft = timeout;
-  HANDLE iocp = processx__get_default_iocp();
   DWORD bytes;
   OVERLAPPED *overlapped = 0;
   ULONG_PTR key;
+  int *events;
 
-  ptr = (int*) R_alloc(npollables, sizeof(int));
+  FD_ZERO(&processx__readfds);
+  FD_ZERO(&processx__writefds);
+  FD_ZERO(&processx__exceptionfds);
 
+  events = (int*) R_alloc(npollables, sizeof(int));
+
+  /* First iteration, we call the pre-poll method, and collect the
+     handles for the IOCP, and the fds for select(). */
   for (i = 0; i < npollables; i++) {
     processx_pollable_t *el = pollables + i;
-    processx_file_handle_t handle = 0;
-    int again;
-    el->event = el->poll_func(el->object, 0, &handle, &again);
-    if (el->event == PXNOPIPE || el->event == PXCLOSED) {
-      /* Do nothing */
-    } else if (el->event == PXREADY) {
-      hasdata++;
-    } else if (el->event == PXSILENT && handle != 0) {
-      ptr[j] = i;
+    events[i] = PXSILENT;
+    if (el->pre_poll_func) events[i] = el->pre_poll_func(el);
+    switch (events[i]) {
+    case PXHANDLE:
       j++;
-    } else {
-      error("Cannnot poll pollable, not ready, and no handle");
+      break;
+    default:
+      break;
     }
   }
 
-  if (j == 0) return hasdata;
+  /* j contains the number of IOCP handles to poll */
+
+  ptr = (int*) R_alloc(j, sizeof(int));
+
+  for (i = 0, j = 0; i < npollables; i++) {
+    processx_pollable_t *el = pollables + i;
+
+    switch (events[i]) {
+    case PXNOPIPE:
+    case PXCLOSED:
+    case PXSILENT:
+      el->event = events[i];
+      break;
+
+    case PXREADY:
+      hasdata++;
+      el->event = events[i];
+      break;
+
+    case PXHANDLE:
+      el->event = PXSILENT;
+      ptr[j] = i;
+      j++;
+      break;
+
+    case PXSELECT: {
+      SEXP elem;
+      el->event = PXSILENT;
+      int k, n;
+      elem = VECTOR_ELT(el->fds, 0);
+      n = LENGTH(elem);
+      selj += n;
+      for (k = 0; k < n; k++) FD_SET(INTEGER(elem)[k], &processx__readfds);
+      elem = VECTOR_ELT(el->fds, 1);
+      n = LENGTH(elem);
+      selj += n;
+      for (k = 0; k < n; k++) FD_SET(INTEGER(elem)[k], &processx__writefds);
+      elem = VECTOR_ELT(el->fds, 2);
+      n = LENGTH(elem);
+      selj += n;
+      for (k = 0; k < n; k++) FD_SET(INTEGER(elem)[k], &processx__exceptionfds);
+    } }
+  }
+
+  if (j == 0 && selj == 0) return hasdata;
 
   if (hasdata) timeout = timeleft = 0;
 
+  if (selj != 0) {
+    processx__socket_pair(processx__notify_socket);
+    FD_SET(processx__notify_socket[0], &processx__readfds);
+    processx__select = 1;
+  } else {
+    processx__select = 0;
+  }
+
   while (timeout < 0 || timeleft >= 0) {
     int poll_timeout;
     if (timeout < 0 || timeleft > PROCESSX_INTERRUPT_INTERVAL) {
@@ -417,12 +759,52 @@ int processx_c_connection_poll(processx_pollable_t pollables[],
       poll_timeout = timeleft;
     }
 
-    GetQueuedCompletionStatus(
-      /* CompletionPort  = */ iocp,
-      /* lpNumberOfBytes = */ &bytes,
-      /* lpCompletionKey = */ &key,
-      /* lpOverlapped    = */ &overlapped,
-      /* dwMilliseconds  = */ poll_timeout);
+    BOOL sres;
+    if (selj == 0) {
+      sres = processx__thread_getstatus(&bytes, &key, &overlapped,
+					poll_timeout);
+    } else {
+      sres = processx__thread_getstatus_select(&bytes, &key, &overlapped,
+					       poll_timeout);
+    }
+    DWORD err = sres ? ERROR_SUCCESS : processx__thread_get_last_error();
+
+    /* See if there was any data from the curl sockets */
+
+    if (processx__select) {
+      for (i = 0; i < npollables; i++) {
+	if (events[i] == PXSELECT) {
+	  processx_pollable_t *el = pollables + i;
+	  SEXP elem;
+	  int k, n;
+	  int has = 0;
+	  elem = VECTOR_ELT(el->fds, 0);
+	  n = LENGTH(elem);
+	  for (k = 0; k < n; k++) {
+	    if (FD_ISSET(INTEGER(elem)[k], &processx__readfds)) has = 1;
+	    FD_SET(INTEGER(elem)[k], &processx__readfds);
+	  }
+	  elem = VECTOR_ELT(el->fds, 1);
+	  n = LENGTH(elem);
+	  for (k = 0; k < n; k++) {
+	    if (FD_ISSET(INTEGER(elem)[k], &processx__writefds)) has = 1;
+	    FD_SET(INTEGER(elem)[k], &processx__writefds);
+	  }
+	  elem = VECTOR_ELT(el->fds, 2);
+	  n = LENGTH(elem);
+	  for (k = 0; k < n; k++) {
+	    if (FD_ISSET(INTEGER(elem)[k], &processx__exceptionfds)) has = 1;
+	    FD_SET(INTEGER(elem)[k], &processx__exceptionfds);
+	  }
+	  if (has) {
+	    el->event = PXEVENT;
+	    hasdata++;
+	  }
+	}
+      }
+    }
+
+    /* See if there was any data from the IOCP */
 
     if (overlapped) {
       /* data */
@@ -432,7 +814,7 @@ int processx_c_connection_poll(processx_pollable_t pollables[],
       con->buffer_data_size += bytes;
       if (con->buffer_data_size > 0) processx__connection_to_utf8(con);
       if (con->type == PROCESSX_FILE_TYPE_ASYNCFILE) {
-	/* TODO: larget files */
+	/* TODO: larger files */
 	con->handle.overlapped.Offset += bytes;
       }
 
@@ -443,17 +825,18 @@ int processx_c_connection_poll(processx_pollable_t pollables[],
 	}
       }
 
+      if (con->handle.freelist) processx__connection_freelist_remove(con);
+
       if (poll_idx < npollables &&
 	  pollables[poll_idx].object == con) {
 	pollables[poll_idx].event = PXREADY;
 	hasdata++;
-	break;
       }
-
-    } else if (GetLastError() != WAIT_TIMEOUT) {
-      PROCESSX_ERROR("Cannot poll", GetLastError());
+    } else if (err != WAIT_TIMEOUT && err != ERROR_SUCCESS) {
+      PROCESSX_ERROR("Cannot poll", err);
     }
 
+    if (hasdata) break;
     R_CheckUserInterrupt();
     timeleft -= PROCESSX_INTERRUPT_INTERVAL;
   }
@@ -462,6 +845,9 @@ int processx_c_connection_poll(processx_pollable_t pollables[],
     for (i = 0; i < j; i++) pollables[ptr[i]].event = PXTIMEOUT;
   }
 
+  closesocket(processx__notify_socket[0]);
+  closesocket(processx__notify_socket[1]);
+
   return hasdata;
 }
 
@@ -469,7 +855,7 @@ int processx_c_connection_poll(processx_pollable_t pollables[],
 
 static int processx__poll_decode(short code) {
   if (code & POLLNVAL) return PXCLOSED;
-  if (code & POLLIN || code & POLLHUP) return PXREADY;
+  if (code & POLLIN || code & POLLHUP || code & POLLOUT) return PXREADY;
   return PXSILENT;
 }
 
@@ -482,36 +868,82 @@ int processx_c_connection_poll(processx_pollable_t pollables[],
   struct pollfd *fds;
   int *ptr;
   int ret;
+  int *events;
 
   if (npollables == 0) return 0;
 
   /* Need to allocate this, because we need to put in the fds, maybe */
-  ptr = (int*) R_alloc(npollables, sizeof(int));
-  fds = (struct pollfd*) R_alloc(npollables, sizeof(struct pollfd));
+  events = (int*) R_alloc(npollables, sizeof(int));
 
-  /* Need to call the poll method for every pollable */
+  /* First iteration, we call the pre-poll method, and collect the
+     fds to poll. */
   for (i = 0; i < npollables; i++) {
     processx_pollable_t *el = pollables + i;
-    processx_file_handle_t handle;
-    int again;
-    el->event = el->poll_func(el->object, 0, &handle, &again);
-    if (el->event == PXNOPIPE || el->event == PXCLOSED) {
-      /* Do nothing */
-    } else if (el->event == PXREADY) {
-      hasdata++;
-    } else if (el->event == PXSILENT && handle >= 0) {
-      fds[j].fd = handle;
-      fds[j].events = POLLIN;
-      fds[j].revents = 0;
-      ptr[j] = (int) i;
+    events[i] = PXSILENT;
+    if (el->pre_poll_func) events[i] = el->pre_poll_func(el);
+    switch (events[i]) {
+    case PXHANDLE:
       j++;
-    } else {
-      error("Cannot poll pollable: not ready and no fd");
+      break;
+    case PXSELECT: {
+      /* This is three vectors of fds to poll, in an R list */
+      int w;
+      for (w = 0; w < 3; w++) {
+        j += LENGTH(VECTOR_ELT(el->fds, w));
+      } }
+    default:
+      break;
     }
   }
 
   /* j contains the number of fds to poll now */
 
+  fds = (struct pollfd*) R_alloc(j, sizeof(struct pollfd));
+  ptr = (int*) R_alloc(j, sizeof(int));
+
+  /* Need to go over them again, collect the ones that we need to poll */
+  for (i = 0, j = 0; i < npollables; i++) {
+    processx_pollable_t *el = pollables + i;
+    switch (events[i]) {
+    case PXNOPIPE:
+    case PXCLOSED:
+    case PXSILENT:
+      el->event = events[i];
+      break;
+
+    case PXREADY:
+      hasdata++;
+      el->event = events[i];
+      break;
+
+    case PXHANDLE:
+      el->event = PXSILENT;
+      fds[j].fd = el->handle;
+      fds[j].events = POLLIN;
+      fds[j].revents = 0;
+      ptr[j] = (int) i;
+      j++;
+      break;
+
+    case PXSELECT: {
+      int pollevs[3] = { POLLIN, POLLOUT, POLLIN | POLLOUT };
+      int w;
+      el->event = PXSILENT;
+      for (w = 0; w < 3; w++) {
+        SEXP elem = VECTOR_ELT(el->fds, w);
+        int k, n = LENGTH(elem);
+        for (k = 0; k < n; k++) {
+          fds[j].fd = INTEGER(elem)[k];
+          fds[j].events = pollevs[w];
+          fds[j].revents = 0;
+          ptr[j] = (int) i;
+          j++;
+        }
+      }
+      break; }
+    }
+  }
+
   /* Nothing to poll */
   if (j == 0) return hasdata;
 
@@ -530,8 +962,17 @@ int processx_c_connection_poll(processx_pollable_t pollables[],
 
   } else {
     for (i = 0; i < j; i++) {
-      pollables[ptr[i]].event = processx__poll_decode(fds[i].revents);
-      hasdata += (pollables[ptr[i]].event == PXREADY);
+      if (events[ptr[i]] == PXSELECT) {
+        if (pollables[ptr[i]].event == PXSILENT) {
+          int ev = fds[i].revents;
+          if (ev & (POLLNVAL | POLLIN | POLLHUP | POLLOUT)) {
+            pollables[ptr[i]].event = PXEVENT;
+          }
+        }
+      } else {
+        pollables[ptr[i]].event = processx__poll_decode(fds[i].revents);
+        hasdata += (pollables[ptr[i]].event == PXREADY);
+      }
     }
   }
 
@@ -555,20 +996,14 @@ void processx__connection_start_read(processx_connection_t *ccon) {
 
   todo = ccon->buffer_allocated_size - ccon->buffer_data_size;
 
-  /* These need to be set to zero for non-file handles */
-  if (ccon->type != PROCESSX_FILE_TYPE_ASYNCFILE) {
-     ccon->handle.overlapped.Offset = 0;
-     ccon->handle.overlapped.OffsetHigh = 0;
-  }
-  res = ReadFile(
-    /* hfile = */                ccon->handle.handle,
-    /* lpBuffer = */             ccon->buffer + ccon->buffer_data_size,
-    /* nNumberOfBytesToRead = */ todo,
-    /* lpNumberOfBytesRead = */  &bytes_read,
-    /* lpOverlapped = */         &ccon->handle.overlapped);
+  res = processx__thread_readfile(
+    ccon,
+    ccon->buffer + ccon->buffer_data_size,
+    todo,
+    &bytes_read);
 
   if (!res) {
-    DWORD err = GetLastError();
+    DWORD err = processx__thread_get_last_error();
     if (err == ERROR_BROKEN_PIPE || err == ERROR_HANDLE_EOF) {
       ccon->is_eof_raw_ = 1;
       if (ccon->utf8_data_size == 0 && ccon->buffer_data_size == 0) {
@@ -606,7 +1041,7 @@ void processx__connection_start_read(processx_connection_t *ccon) {
  *    to convert it to UTF8.
  */
 
-#define PROCESSX__I_POLL_FUNC_CONNECTION_READY do {			\
+#define PROCESSX__I_PRE_POLL_FUNC_CONNECTION_READY do {			\
   if (!ccon) return PXNOPIPE;						\
   if (ccon->is_closed_) return PXCLOSED;				\
   if (ccon->is_eof_) return PXREADY;					\
@@ -617,40 +1052,59 @@ void processx__connection_start_read(processx_connection_t *ccon) {
     if (ccon->utf8_data_size > 0) return PXREADY;			\
   } } while (0)
 
-int processx_i_poll_func_connection(
-  void * object,
-  int status,
-  processx_file_handle_t *handle,
-  int *again) {
+int processx_i_pre_poll_func_connection(processx_pollable_t *pollable) {
 
-  processx_connection_t *ccon = (processx_connection_t*) object;
+  processx_connection_t *ccon = pollable->object;
 
-  PROCESSX__I_POLL_FUNC_CONNECTION_READY;
+  PROCESSX__I_PRE_POLL_FUNC_CONNECTION_READY;
 
 #ifdef _WIN32
   processx__connection_start_read(ccon);
   /* Starting to read may actually get some data, or an EOF, so check again */
-  PROCESSX__I_POLL_FUNC_CONNECTION_READY;
-  if (handle) *handle = ccon->handle.overlapped.hEvent;
+  PROCESSX__I_PRE_POLL_FUNC_CONNECTION_READY;
+  pollable->handle = ccon->handle.overlapped.hEvent;
 #else
-  if (handle) *handle = ccon->handle;
+  pollable->handle = ccon->handle;
 #endif
 
-  if (again) *again = 0;
-
-  return PXSILENT;
+  return PXHANDLE;
 }
 
 int processx_c_pollable_from_connection(
   processx_pollable_t *pollable,
   processx_connection_t *ccon) {
 
-  pollable->poll_func = processx_i_poll_func_connection;
+  pollable->pre_poll_func = processx_i_pre_poll_func_connection;
   pollable->object = ccon;
   pollable->free = 0;
+  pollable->fds = R_NilValue;
   return 0;
 }
 
+int processx_i_pre_poll_func_curl(processx_pollable_t *pollable) {
+  return PXSELECT;
+}
+
+int processx_c_pollable_from_curl(
+  processx_pollable_t *pollable,
+  SEXP fds) {
+
+  pollable->pre_poll_func = processx_i_pre_poll_func_curl;
+  pollable->object = NULL;
+  pollable->free = 0;
+  pollable->fds = fds;
+  return 0;
+}
+
+processx_file_handle_t processx_c_connection_fileno(
+  const processx_connection_t *con) {
+#ifdef _WIN32
+  return con->handle.handle;
+#else
+  return con->handle;
+#endif
+}
+
 /* --------------------------------------------------------------------- */
 /* Internals                                                             */
 /* --------------------------------------------------------------------- */
@@ -738,7 +1192,6 @@ static void processx__connection_find_lines(processx_connection_t *ccon,
 
 static void processx__connection_xfinalizer(SEXP con) {
   processx_connection_t *ccon = R_ExternalPtrAddr(con);
-
   processx_c_connection_destroy(ccon);
 }
 
@@ -834,7 +1287,7 @@ static void processx__connection_realloc(processx_connection_t *ccon) {
 
 #ifdef _WIN32
 
-static ssize_t processx__connection_read(processx_connection_t *ccon) {
+ssize_t processx__connection_read(processx_connection_t *ccon) {
   DWORD todo, bytes_read = 0;
 
   /* Nothing to read, nothing to convert to UTF8 */
@@ -854,19 +1307,13 @@ static ssize_t processx__connection_read(processx_connection_t *ccon) {
 
   /* A read might be pending at this point. See if it has finished. */
   if (ccon->handle.read_pending) {
-    HANDLE iocp = processx__get_default_iocp();
     ULONG_PTR key;
     DWORD bytes;
     OVERLAPPED *overlapped = 0;
 
     while (1) {
-      GetQueuedCompletionStatus(
-        /* CompletionPort  = */ iocp,
-	/* lpNumberOfBytes = */ &bytes,
-	/* lpCompletionKey = */ &key,
-	/* lpOverlapped    = */ &overlapped,
-	/* dwMilliseconds  = */ 0);
-
+      BOOL sres = processx__thread_getstatus(&bytes, &key, &overlapped, 0);
+      DWORD err = sres ? ERROR_SUCCESS : processx__thread_get_last_error();
       if (overlapped) {
 	processx_connection_t *con = (processx_connection_t *) key;
 	con->handle.read_pending = FALSE;
@@ -885,13 +1332,15 @@ static ssize_t processx__connection_read(processx_connection_t *ccon) {
 	  }
 	}
 
+	if (con->handle.freelist) processx__connection_freelist_remove(con);
+
 	if (con == ccon) {
 	  bytes_read = bytes;
 	  break;
 	}
 
-      } else if (GetLastError() != WAIT_TIMEOUT) {
-	PROCESSX_ERROR("Read error", GetLastError());
+      } else if (err != WAIT_TIMEOUT) {
+	PROCESSX_ERROR("Read error", err);
 
       } else {
 	break;
@@ -1094,4 +1543,51 @@ int processx__interruptible_poll(struct pollfd fds[],
 
 #endif
 
+#ifdef _WIN32
+
+processx__connection_freelist_t freelist_head = { 0, 0 };
+processx__connection_freelist_t *freelist = &freelist_head;
+
+int processx__connection_freelist_add(processx_connection_t *ccon) {
+  if (ccon->handle.freelist) return 0;
+  processx__connection_freelist_t *node =
+    calloc(1, sizeof(processx__connection_freelist_t));
+  if (!node) error("Cannot add to connection freelist, this is a leak");
+
+  node->ccon = ccon;
+  node->next = freelist->next;
+  freelist->next = node;
+  ccon->handle.freelist = TRUE;
+
+  return 0;
+}
+
+void processx__connection_freelist_remove(processx_connection_t *ccon) {
+  processx__connection_freelist_t *prev = freelist, *ptr = freelist->next;
+  while (ptr) {
+    if (ptr->ccon == ccon) {
+      prev->next = ptr->next;
+      free(ptr);
+      return;
+    }
+    prev = ptr;
+    ptr = ptr->next;
+  }
+}
+
+int processx__connection_schedule_destroy(processx_connection_t *ccon) {
+  /* The connection is already closed here, but reads might still be
+     pending... if this is the case, then we add the connection to the
+     free list. */
+  if (ccon->handle.read_pending) {
+    processx__connection_freelist_add(ccon);
+    return 1;
+
+  } else {
+    return 0;
+  }
+}
+
+#endif
+
 #undef PROCESSX_CHECK_VALID_CONN
