@@ -7,15 +7,19 @@
 #define _POSIX_C_SOURCE  200809L
 #endif
 
-#include <Rinternals.h>
-#include <R_ext/Riconv.h>
-
 #ifdef _WIN32
+#ifndef FD_SETSIZE
+#define FD_SETSIZE 32767
+#endif
+#include <winsock2.h>
 #include <windows.h>
 #else
 #include <unistd.h>
 #endif
 
+#include <Rinternals.h>
+#include <R_ext/Riconv.h>
+
 /* --------------------------------------------------------------------- */
 /* Data types                                                            */
 /* --------------------------------------------------------------------- */
@@ -26,9 +30,10 @@
 typedef HANDLE processx_file_handle_t;
 typedef struct {
   HANDLE handle;
-  BOOLEAN async;
   OVERLAPPED overlapped;
+  BOOLEAN async;
   BOOLEAN read_pending;
+  BOOLEAN freelist;
 } processx_i_connection_t;
 #else
 typedef int processx_file_handle_t;
@@ -66,10 +71,11 @@ typedef struct processx_connection_s {
   int poll_idx;
 } processx_connection_t;
 
+struct processx_pollable_s;
+
 /* Generic poll method
  *
  * @param object The thing to poll.
- * @param status Currently not used.
  * @param handle A handle can be returned here, to `poll` or wait on.
  *   If this is not needed, set it to NULL.
  * @param timeout A timeout value can be returned here, for the next
@@ -79,17 +85,16 @@ typedef struct processx_connection_s {
  *     (But maybe not a full line.)
  *   PXSILENT: we don't know if data is available, we need to check the
  *     operating system via `poll` or `WaitForStatus`.
+ *   PXHANDLE
+ *   PXPOLLFD
  */
 
-typedef int (*processx_connection_poll_func_t)(
-  void *object,
-  int status,
-  processx_file_handle_t *handle,
-  int *again);
+typedef int (*processx_connection_pre_poll_func_t)(
+  struct processx_pollable_s *pollable);
 
 /* Data structure for a pollable object
  *
- * @member poll_func The function to call on the object, before
+ * @member pre_poll_func The function to call on the object, before
  *   the poll/wait system call. The pollable object might have data
  *   available without immediately, without poll/wait. If not, it
  *   will return the file descriptor or HANDLE to poll.
@@ -98,13 +103,17 @@ typedef int (*processx_connection_poll_func_t)(
  *   `processx_pollable_t` objects.
  * @member event The result of the polling is stored here. Possible values:
  *   `PXSILENT` (no data), `PXREADY` (data), `PXTIMEOUT` (timeout).
+ * @member fd If the pollable is an fd, then it is stored here instead of
+ *   in `object`, for simplicity.
  */
 
 typedef struct processx_pollable_s {
-  processx_connection_poll_func_t poll_func;
+  processx_connection_pre_poll_func_t pre_poll_func;
   void *object;
   int free;
   int event;
+  processx_file_handle_t handle;
+  SEXP fds;
 } processx_pollable_t;
 
 /* --------------------------------------------------------------------- */
@@ -202,6 +211,9 @@ int processx_c_pollable_from_connection(
   processx_pollable_t *pollable,
   processx_connection_t *ccon);
 
+int processx_c_pollable_from_curl(
+  processx_pollable_t *pollable, SEXP fds);
+
 processx_file_handle_t processx_c_connection_fileno(
   const processx_connection_t *con);
 
@@ -213,13 +225,55 @@ processx_file_handle_t processx_c_connection_fileno(
 typedef unsigned long DWORD;
 #endif
 
+/* Threading in Windows */
+
 #ifdef _WIN32
-extern HANDLE processx__connection_iocp;
-HANDLE processx__get_default_iocp();
+int processx__start_thread();
+extern HANDLE processx__iocp_thread;
+extern HANDLE processx__thread_start;
+extern HANDLE processx__thread_done;
+
+extern fd_set processx__readfds, processx__writefds,
+  processx__exceptionfds;
+extern SOCKET processx__notify_socket[2];
+extern int processx__select;
+extern ULONG_PTR processx__key_none;
+
+extern int processx__thread_cmd;
+#define PROCESSX__THREAD_CMD_INIT 0
+#define PROCESSX__THREAD_CMD_IDLE 1
+#define PROCESSX__THREAD_CMD_READFILE 2
+#define PROCESSX__THREAD_CMD_GETSTATUS 3
+
+BOOL processx__thread_readfile(processx_connection_t *ccon,
+			       LPVOID lpBuffer,
+			       DWORD nNumberOfBytesToRead,
+			       LPDWORD lpNumberOfBytesRead);
+BOOL processx__thread_getstatus(LPDWORD lpNumberOfBytes,
+				PULONG_PTR lpCompletionKey,
+				LPOVERLAPPED *lpOverlapped,
+				DWORD dwMilliseconds);
+BOOL processx__thread_getstatus_select(LPDWORD lpNumberOfBytes,
+				       PULONG_PTR lpCompletionKey,
+				       LPOVERLAPPED *lpOverlapped,
+				       DWORD dwMilliseconds);
+DWORD processx__thread_get_last_error();
+
 #endif
 
 #define PROCESSX_ERROR(m,c) processx__error((m),(c),__FILE__,__LINE__)
 void processx__error(const char *message, DWORD errorcode,
 		     const char *file, int line);
 
+/* Free-list of connection in Windows */
+
+typedef struct processx__connection_freelist_s {
+  processx_connection_t *ccon;
+  struct processx__connection_freelist_s *next;
+} processx__connection_freelist_t;
+
+int processx__connection_freelist_add(processx_connection_t *con);
+void processx__connection_freelist_remove(processx_connection_t *con);
+int processx__connection_schedule_destroy(processx_connection_t *con);
+
 #endif
