@@ -515,7 +515,7 @@ SEXP processx_wait(SEXP status, SEXP timeout) {
 
   if (!handle) {
     processx__unblock_sigchld();
-    error("Internal processx error, handle already removed");
+    return ScalarLogical(1);
   }
 
   pid = handle->pid;
@@ -611,11 +611,7 @@ SEXP processx_is_alive(SEXP status) {
 
   processx__block_sigchld();
 
-  if (!handle) {
-    processx__unblock_sigchld();
-    error("Internal processx error, handle already removed");
-  }
-
+  if (!handle) goto cleanup;
   if (handle->collected) goto cleanup;
 
   /* Otherwise a non-blocking waitpid to collect zombies */
@@ -662,8 +658,8 @@ SEXP processx_get_exit_status(SEXP status) {
   processx__block_sigchld();
 
   if (!handle) {
-    processx__unblock_sigchld();
-    error("Internal processx error, handle already removed");
+    result = PROTECT(ScalarInteger(NA_INTEGER));
+    goto cleanup;
   }
 
   /* If we already have the status, then just return */
@@ -725,8 +721,8 @@ SEXP processx_signal(SEXP status, SEXP signal) {
   processx__block_sigchld();
 
   if (!handle) {
-    processx__unblock_sigchld();
-    error("Internal processx error, handle already removed");
+    result = 0;
+    goto cleanup;
   }
 
   /* If we already have the status, then return `FALSE` */
@@ -792,10 +788,7 @@ SEXP processx_kill(SEXP status, SEXP grace) {
 
   processx__block_sigchld();
 
-  if (!handle) {
-    processx__unblock_sigchld();
-    error("Internal processx error, handle already removed");
-  }
+  if (!handle) { goto cleanup; }
 
   /* Check if we have an exit status, it yes, just return (FALSE) */
   if (handle->collected) { goto cleanup; }
@@ -851,7 +844,9 @@ SEXP processx_kill(SEXP status, SEXP grace) {
 SEXP processx_get_pid(SEXP status) {
   processx_handle_t *handle = R_ExternalPtrAddr(status);
 
-  if (!handle) { error("Internal processx error, handle already removed"); }
+  /* This might happen if it was finalized at the end of the session,
+     even though there are some references to the R object. */
+  if (!handle) return ScalarInteger(NA_INTEGER);
 
   return ScalarInteger(handle->pid);
 }
