@@ -4,6 +4,8 @@
 
 #include "../processx.h"
 
+#include <wchar.h>
+
 static HANDLE processx__global_job_handle = NULL;
 
 static void processx__init_global_job_handle(void) {
@@ -53,7 +55,16 @@ void R_init_processx_win() {
   /* Nothing to do currently */
 }
 
-SEXP processx__killem_all() {
+SEXP processx__unload_cleanup() {
+
+  if (processx__connection_iocp) CloseHandle(processx__connection_iocp);
+  if (processx__iocp_thread) TerminateThread(processx__iocp_thread, 0);
+  if (processx__thread_start) CloseHandle(processx__thread_start);
+  if (processx__thread_done) CloseHandle(processx__thread_done);
+
+  processx__connection_iocp = processx__iocp_thread =
+    processx__thread_start = processx__thread_done = NULL;
+
   if (processx__global_job_handle) {
     TerminateJobObject(processx__global_job_handle, 1);
     CloseHandle(processx__global_job_handle);
@@ -247,6 +258,280 @@ error:
   return err;
 }
 
+/*
+ * The way windows takes environment variables is different than what C does;
+ * Windows wants a contiguous block of null-terminated strings, terminated
+ * with an additional null.
+ *
+ * Windows has a few "essential" environment variables. winsock will fail
+ * to initialize if SYSTEMROOT is not defined; some APIs make reference to
+ * TEMP. SYSTEMDRIVE is probably also important. We therefore ensure that
+ * these get defined if the input environment block does not contain any
+ * values for them.
+ *
+ * Also add variables known to Cygwin to be required for correct
+ * subprocess operation in many cases:
+ * https://github.com/Alexpux/Cygwin/blob/b266b04fbbd3a595f02ea149e4306d3ab9b1fe3d/winsup/cygwin/environ.cc#L955
+ *
+ */
+
+typedef struct env_var {
+  const WCHAR* const wide;
+  const WCHAR* const wide_eq;
+  const size_t len; /* including null or '=' */
+} env_var_t;
+
+#define E_V(str) { L##str, L##str L"=", sizeof(str) }
+
+static const env_var_t required_vars[] = { /* keep me sorted */
+  E_V("HOMEDRIVE"),
+  E_V("HOMEPATH"),
+  E_V("LOGONSERVER"),
+  E_V("PATH"),
+  E_V("SYSTEMDRIVE"),
+  E_V("SYSTEMROOT"),
+  E_V("TEMP"),
+  E_V("USERDOMAIN"),
+  E_V("USERNAME"),
+  E_V("USERPROFILE"),
+  E_V("WINDIR"),
+};
+static size_t n_required_vars = ARRAY_SIZE(required_vars);
+
+int env_strncmp(const wchar_t* a, int na, const wchar_t* b) {
+  wchar_t* a_eq;
+  wchar_t* b_eq;
+  wchar_t* A;
+  wchar_t* B;
+  int nb;
+  int r;
+
+  if (na < 0) {
+    a_eq = wcschr(a, L'=');
+    na = (int)(long)(a_eq - a);
+  } else {
+    na--;
+  }
+  b_eq = wcschr(b, L'=');
+  nb = b_eq - b;
+
+  A = alloca((na+1) * sizeof(wchar_t));
+  B = alloca((nb+1) * sizeof(wchar_t));
+
+  r = LCMapStringW(LOCALE_INVARIANT, LCMAP_UPPERCASE, a, na, A, na);
+  if (!r) PROCESSX_ERROR("make environment", GetLastError());
+  A[na] = L'\0';
+  r = LCMapStringW(LOCALE_INVARIANT, LCMAP_UPPERCASE, b, nb, B, nb);
+  if (!r) PROCESSX_ERROR("make environment", GetLastError());
+  B[nb] = L'\0';
+
+  while (1) {
+    wchar_t AA = *A++;
+    wchar_t BB = *B++;
+    if (AA < BB) {
+      return -1;
+    } else if (AA > BB) {
+      return 1;
+    } else if (!AA && !BB) {
+      return 0;
+    }
+  }
+}
+
+static int qsort_wcscmp(const void *a, const void *b) {
+  wchar_t* astr = *(wchar_t* const*)a;
+  wchar_t* bstr = *(wchar_t* const*)b;
+  return env_strncmp(astr, -1, bstr);
+}
+
+static int processx__add_tree_id_env(const char *ctree_id, WCHAR **dst_ptr) {
+  WCHAR *env = GetEnvironmentStringsW();
+  int len = 0, len2 = 0;
+  WCHAR *ptr = env;
+  WCHAR *id = 0;
+  int err;
+  int idlen;
+  WCHAR *dst_copy;
+
+  if (!env) return GetLastError();
+
+  err = processx__utf8_to_utf16_alloc(ctree_id, &id);
+  if (err) {
+    FreeEnvironmentStringsW(env);
+    return(err);
+  }
+
+  while (1) {
+    WCHAR *prev = ptr;
+    if (!*ptr) break;
+    while (*ptr) ptr++;
+    ptr++;
+    len += (ptr - prev);
+  }
+
+  /* Plus the id */
+  idlen = wcslen(id) + 1;
+  len2 = len + idlen;
+
+  /* Allocate, copy */
+  dst_copy = (WCHAR*) R_alloc(len2 + 1, sizeof(WCHAR)); /* +1 for final zero */
+  memcpy(dst_copy, env, len * sizeof(WCHAR));
+  memcpy(dst_copy + len, id, idlen * sizeof(WCHAR));
+
+  /* Final \0 */
+  *(dst_copy + len2) = L'\0';
+  *dst_ptr = dst_copy;
+
+  FreeEnvironmentStringsW(env);
+  return 0;
+}
+
+static int processx__make_program_env(SEXP env_block, const char *tree_id, WCHAR** dst_ptr) {
+  WCHAR* dst;
+  WCHAR* ptr;
+  size_t env_len = 0;
+  int len;
+  size_t i;
+  DWORD var_size;
+  size_t env_block_count = 1; /* 1 for null-terminator */
+  WCHAR* dst_copy;
+  WCHAR** ptr_copy;
+  WCHAR** env_copy;
+  DWORD* required_vars_value_len = alloca(n_required_vars * sizeof(DWORD*));
+  int j, num = LENGTH(env_block);
+
+  /* first pass: determine size in UTF-16 */
+  for (j = 0; j < num; j++) {
+    const char *env = CHAR(STRING_ELT(env_block, 0));
+    if (strchr(env, '=')) {
+      len = MultiByteToWideChar(CP_UTF8,
+                                0,
+                                env,
+                                -1,
+                                NULL,
+                                0);
+      if (len <= 0) {
+        return GetLastError();
+      }
+      env_len += len;
+      env_block_count++;
+    }
+  }
+
+  /* Plus the tree id */
+  len = MultiByteToWideChar(CP_UTF8, 0, tree_id, -1, NULL, 0);
+  if (len <= 0) return GetLastError();
+  env_len += len;
+  env_block_count++;
+
+  /* second pass: copy to UTF-16 environment block */
+  dst_copy = (WCHAR*) R_alloc(env_len, sizeof(WCHAR));
+  env_copy = alloca(env_block_count * sizeof(WCHAR*));
+
+  ptr = dst_copy;
+  ptr_copy = env_copy;
+  for (j = 0; j < num; j++) {
+    const char *env = CHAR(STRING_ELT(env_block, 0));
+    if (strchr(env, '=')) {
+      len = MultiByteToWideChar(CP_UTF8,
+                                0,
+                                env,
+                                -1,
+                                ptr,
+                                (int) (env_len - (ptr - dst_copy)));
+      if (len <= 0) {
+        DWORD err = GetLastError();
+        return err;
+      }
+      *ptr_copy++ = ptr;
+      ptr += len;
+    }
+  }
+  /* Plus the tree id */
+  len = MultiByteToWideChar(CP_UTF8, 0, tree_id, -1, ptr,
+			    (int) (env_len  - (ptr - dst_copy)));
+  if (len <= 0) return GetLastError();
+  *ptr_copy++ = ptr;
+  ptr += len;
+
+  *ptr_copy = NULL;
+
+  /* sort our (UTF-16) copy */
+  qsort(env_copy, env_block_count-1, sizeof(wchar_t*), qsort_wcscmp);
+
+  /* third pass: check for required variables */
+  for (ptr_copy = env_copy, i = 0; i < n_required_vars; ) {
+    int cmp;
+    if (!*ptr_copy) {
+      cmp = -1;
+    } else {
+      cmp = env_strncmp(required_vars[i].wide_eq,
+                       required_vars[i].len,
+                        *ptr_copy);
+    }
+    if (cmp < 0) {
+      /* missing required var */
+      var_size = GetEnvironmentVariableW(required_vars[i].wide, NULL, 0);
+      required_vars_value_len[i] = var_size;
+      if (var_size != 0) {
+        env_len += required_vars[i].len;
+        env_len += var_size;
+      }
+      i++;
+    } else {
+      ptr_copy++;
+      if (cmp == 0)
+        i++;
+    }
+  }
+
+  /* final pass: copy, in sort order, and inserting required variables */
+  dst = (WCHAR*) R_alloc(1 + env_len, sizeof(WCHAR));
+
+  for (ptr = dst, ptr_copy = env_copy, i = 0;
+       *ptr_copy || i < n_required_vars;
+       ptr += len) {
+    int cmp;
+    if (i >= n_required_vars) {
+      cmp = 1;
+    } else if (!*ptr_copy) {
+      cmp = -1;
+    } else {
+      cmp = env_strncmp(required_vars[i].wide_eq,
+                        required_vars[i].len,
+                        *ptr_copy);
+    }
+    if (cmp < 0) {
+      /* missing required var */
+      len = required_vars_value_len[i];
+      if (len) {
+        wcscpy(ptr, required_vars[i].wide_eq);
+        ptr += required_vars[i].len;
+        var_size = GetEnvironmentVariableW(required_vars[i].wide,
+                                           ptr,
+                                           (int) (env_len - (ptr - dst)));
+        if (var_size != len-1) { /* race condition? */
+          PROCESSX_ERROR("GetEnvironmentVariableW", GetLastError());
+        }
+      }
+      i++;
+    } else {
+      /* copy var from env_block */
+      len = wcslen(*ptr_copy) + 1;
+      wmemcpy(ptr, *ptr_copy, len);
+      ptr_copy++;
+      if (cmp == 0)
+        i++;
+    }
+  }
+
+  /* Terminate with an extra NULL. */
+  *ptr = L'\0';
+
+  *dst_ptr = dst;
+  return 0;
+}
+
 static WCHAR* processx__search_path_join_test(const WCHAR* dir,
 					      size_t dir_len,
 					      const WCHAR* name,
@@ -563,14 +848,8 @@ DWORD processx__terminate(processx_handle_t *handle, SEXP status) {
   return err;
 }
 
-SEXP processx__disconnect_process_handle(SEXP status) {
-  R_SetExternalPtrTag(status, R_NilValue);
-  return R_NilValue;
-}
-
 void processx__finalizer(SEXP status) {
   processx_handle_t *handle = (processx_handle_t*) R_ExternalPtrAddr(status);
-  SEXP private;
 
   if (!handle) return;
 
@@ -579,21 +858,8 @@ void processx__finalizer(SEXP status) {
     processx__terminate(handle, status);
   }
 
-  /* Copy over pid and exit status */
-  private = PROTECT(R_ExternalPtrTag(status));
-  if (!isNull(private)) {
-    SEXP sone = PROTECT(ScalarLogical(1));
-    SEXP spid = PROTECT(ScalarInteger(handle->dwProcessId));
-    SEXP sexitcode = PROTECT(ScalarInteger(handle->exitcode));
-
-    defineVar(install("exited"), sone, private);
-    defineVar(install("pid"), spid, private);
-    defineVar(install("exitcode"), sexitcode, private);
-    UNPROTECT(3);
-  }
-  UNPROTECT(1);
-
   if (handle->hProcess) CloseHandle(handle->hProcess);
+  handle->hProcess = NULL;
   R_ClearExternalPtr(status);
   processx__handle_destroy(handle);
 }
@@ -620,19 +886,27 @@ void processx__handle_destroy(processx_handle_t *handle) {
   free(handle);
 }
 
-SEXP processx_exec(SEXP command, SEXP args, SEXP std_out, SEXP std_err,
+SEXP processx_exec(SEXP command, SEXP args,
+		   SEXP std_in, SEXP std_out, SEXP std_err,
+		   SEXP connections, SEXP env,
 		   SEXP windows_verbatim_args, SEXP windows_hide,
-		   SEXP private, SEXP cleanup, SEXP wd, SEXP encoding) {
+		   SEXP private, SEXP cleanup, SEXP wd, SEXP encoding,
+		   SEXP tree_id) {
 
+  const char *cstd_in = isNull(std_in) ? 0 : CHAR(STRING_ELT(std_in, 0));
   const char *cstd_out = isNull(std_out) ? 0 : CHAR(STRING_ELT(std_out, 0));
   const char *cstd_err = isNull(std_err) ? 0 : CHAR(STRING_ELT(std_err, 0));
   const char *cencoding = CHAR(STRING_ELT(encoding, 0));
   const char *ccwd = isNull(wd) ? 0 : CHAR(STRING_ELT(wd, 0));
+  const char *ctree_id = CHAR(STRING_ELT(tree_id, 0));
+  int i, num_connections = LENGTH(connections) + 3;
+  HANDLE* extra_connections =
+    (HANDLE*) R_alloc(num_connections - 3, sizeof(HANDLE));
 
   int err = 0;
   WCHAR *path;
   WCHAR *application_path = NULL, *application = NULL, *arguments = NULL,
-    *cwd = NULL;
+    *cenv = NULL, *cwd = NULL;
   processx_options_t options;
   STARTUPINFOW startup = { 0 };
   PROCESS_INFORMATION info = { 0 };
@@ -656,6 +930,13 @@ SEXP processx_exec(SEXP command, SEXP args, SEXP std_out, SEXP std_err,
       &arguments);
   if (err) { PROCESSX_ERROR("making program args", err); }
 
+  if (isNull(env)) {
+    err = processx__add_tree_id_env(ctree_id, &cenv);
+  } else {
+    err = processx__make_program_env(env, ctree_id, &cenv);
+  }
+  if (err) PROCESSX_ERROR("making environment", err);
+
   if (ccwd) {
     /* Explicit cwd */
     err = processx__utf8_to_utf16_alloc(ccwd, &cwd);
@@ -700,7 +981,14 @@ SEXP processx_exec(SEXP command, SEXP args, SEXP std_out, SEXP std_err,
   result = PROTECT(processx__make_handle(private, ccleanup));
   handle = R_ExternalPtrAddr(result);
 
-  err = processx__stdio_create(handle, cstd_out, cstd_err,
+  for (i = 0; i < num_connections - 3; i++) {
+    processx_connection_t *ccon =
+      R_ExternalPtrAddr(VECTOR_ELT(connections, i));
+    extra_connections[i] = processx_c_connection_fileno(ccon);
+  }
+
+  err = processx__stdio_create(handle, extra_connections, num_connections,
+			       cstd_in, cstd_out, cstd_err,
 			       &handle->child_stdio_buffer, private,
 			       cencoding);
   if (err) { PROCESSX_ERROR("setup stdio", err); }
@@ -753,7 +1041,7 @@ SEXP processx_exec(SEXP command, SEXP args, SEXP std_out, SEXP std_err,
     /* lpThreadAttributes =   */ NULL,
     /* bInheritHandles =      */ 1,
     /* dwCreationFlags =      */ process_flags,
-    /* lpEnvironment =        */ NULL,
+    /* lpEnvironment =        */ cenv,
     /* lpCurrentDirectory =   */ cwd,
     /* lpStartupInfo =        */ &startup,
     /* lpProcessInformation = */ &info);
@@ -763,9 +1051,14 @@ SEXP processx_exec(SEXP command, SEXP args, SEXP std_out, SEXP std_err,
   handle->hProcess = info.hProcess;
   handle->dwProcessId = info.dwProcessId;
 
+  /* Query official creation time. On Windows this is not used as
+     an id, since the pid itself is valid until the process handle
+     is released. */
+  handle->create_time = processx__create_time(handle->hProcess);
+
   /* If the process isn't spawned as detached, assign to the global job */
   /* object so windows will kill it when the parent process dies. */
-  if (!ccleanup) {
+  if (ccleanup) {
     if (! processx__global_job_handle) processx__init_global_job_handle();
 
     if (!AssignProcessToJobObject(processx__global_job_handle, info.hProcess)) {
@@ -808,7 +1101,8 @@ SEXP processx_wait(SEXP status, SEXP timeout) {
   processx_handle_t *handle = R_ExternalPtrAddr(status);
   DWORD err, err2, exitcode;
 
-  if (handle->collected) return R_NilValue;
+  if (!handle) return ScalarLogical(1);
+  if (handle->collected) return ScalarLogical(1);
 
   err2 = WAIT_TIMEOUT;
   while (ctimeout < 0 || timeleft > PROCESSX_INTERRUPT_INTERVAL) {
@@ -842,6 +1136,10 @@ SEXP processx_is_alive(SEXP status) {
   processx_handle_t *handle = R_ExternalPtrAddr(status);
   DWORD err, exitcode;
 
+  /* This might happen if it was finalized at the end of the session,
+     even though there are some references to the R object. */
+  if (!handle) return ScalarLogical(0);
+
   if (handle->collected) return ScalarLogical(0);
 
   /* Otherwise try to get exit code */
@@ -862,6 +1160,10 @@ SEXP processx_get_exit_status(SEXP status) {
   processx_handle_t *handle = R_ExternalPtrAddr(status);
   DWORD err, exitcode;
 
+  /* This might happen if it was finalized at the end of the session,
+     even though there are some references to the R object. */
+  if (!handle) return R_NilValue;
+
   if (handle->collected) return ScalarInteger(handle->exitcode);
 
   /* Otherwise try to get exit code */
@@ -880,6 +1182,7 @@ SEXP processx_signal(SEXP status, SEXP signal) {
   processx_handle_t *handle = R_ExternalPtrAddr(status);
   DWORD err, exitcode = STILL_ACTIVE;
 
+  if (!handle) return ScalarLogical(0);
   if (handle->collected) return ScalarLogical(0);
 
   switch (INTEGER(signal)[0]) {
@@ -896,11 +1199,6 @@ SEXP processx_signal(SEXP status, SEXP signal) {
     }
 
     if (exitcode == STILL_ACTIVE) {
-
-      /* We only cleanup the tree if the process is still running. */
-      /* TODO: we are not running this for now, until we can properly
-	 work around pid reuse. */
-      /* processx__cleanup_child_tree(handle->dwProcessId); */
       err = processx__terminate(handle, status);
       return ScalarLogical(err != 0);
 
@@ -930,6 +1228,11 @@ SEXP processx_signal(SEXP status, SEXP signal) {
   }
 }
 
+SEXP processx_interrupt(SEXP status) {
+  error("Internal processx error, `processx_interrupt()` should not be called");
+  return R_NilValue;
+}
+
 SEXP processx_kill(SEXP status, SEXP grace) {
   return processx_signal(status, ScalarInteger(9));
 }
@@ -937,7 +1240,9 @@ SEXP processx_kill(SEXP status, SEXP grace) {
 SEXP processx_get_pid(SEXP status) {
   processx_handle_t *handle = R_ExternalPtrAddr(status);
 
-  if (!handle) { error("Internal processx error, handle already removed"); }
+  /* This might happen if it was finalized at the end of the session,
+     even though there are some references to the R object. */
+  if (!handle) return ScalarInteger(NA_INTEGER);
 
   return ScalarInteger(handle->dwProcessId);
 }
