@@ -3,6 +3,19 @@ skip_other_platforms <- function(platform) {
   if (os_type() != platform) skip(paste("only run it on", platform))
 }
 
+skip_if_no_tool <- function(tool) {
+  if (Sys.which(tool) == "") skip(paste0("`", tool, "` is not available"))
+}
+
+skip_extra_tests <- function() {
+  if (Sys.getenv("PROCESSX_EXTRA_TESTS") ==  "") skip("no extra tests")
+}
+
+skip_if_no_ps <- function() {
+  if (!requireNamespace("ps", quietly = TRUE)) skip("ps package needed")
+  if (!ps::ps_is_supported()) skip("ps does not support this platform")
+}
+
 try_silently <- function(expr) {
   tryCatch(
     expr,
@@ -12,18 +25,6 @@ try_silently <- function(expr) {
   )
 }
 
-get_tool <- function(prog) {
-  if (os_type() == "windows") prog <- paste0(prog, ".exe")
-  exe <- system.file(package = "processx", "bin", .Platform$r_arch, prog)
-  if (exe == "") {
-    pkgpath <- system.file(package = "processx")
-    if (basename(pkgpath) == "inst") pkgpath <- dirname(pkgpath)
-    exe <- file.path(pkgpath, "src", "tools", prog)
-    if (!file.exists(exe)) return("")
-  }
-  exe
-}
-
 get_pid_by_name <- function(name) {
   if (os_type() == "windows") {
     get_pid_by_name_windows(name)
@@ -54,3 +55,15 @@ get_pid_by_name_windows <- function(name) {
 get_pid_by_name_linux <- function(name) {
   ## TODO
 }
+
+skip_in_covr <- function() {
+  if (Sys.getenv("R_COVR", "") == "true") skip("in covr")
+}
+
+httpbin <- local({
+  cache <- NULL
+  function(url = "") {
+    if (is.null(cache)) cache <<- curl::nslookup("eu.httpbin.org")
+    paste0("http://", cache, url)
+  }
+})
